__author__ = 'Jacob C. Kimmel, David R. Kelley'
__email__ = 'jacobkimmel+scnym@gmail.com, drk@calicolabs.com'
__version__ = '0.3.1'

# populate the namespace so top level imports work
# e.g.
# >> from scnym.model import CellTypeCLF
from . import api, main, dataprep, interpret, model, predict, trainer, utils

